//-----------------------------------------------------------------------------
// File:     ProjectileMotion.cpp
// Class:    None
// Parent:   None
// Children: None
// Purpose:  Simulates a baseball being hit
//-----------------------------------------------------------------------------
// The following are standard C/C++ header files.
// If a filename is enclosed inside < >  it means the header file is in the Include directory.
// If a filename is enclosed inside " "  it means the header file is in the current directory.
#include <ctype.h>      // Character Types
#include <math.h>       // Mathematical Constants
#include <stdarg.h>     // Variable Argument Lists
#include <stdio.h>      // Standard Input/Output Functions
#include <stdlib.h>     // Utility Functions
#include <string.h>     // String Operations
#include <signal.h>     // Signals (Contol-C + Unix System Calls)
#include <setjmp.h>     // Nonlocal Goto (For Control-C)
#include <time.h>       // Time and Date information
#include <assert.h>     // Verify Program Assertion
#include <errno.h>      // Error Codes (Used in Unix system())
#include <float.h>      // Floating Point Constants
#include <limits.h>     // Implementation Constants
#include <stddef.h>     // Standard Definitions
#include <exception>    // Exception handling (e.g., try, catch throw)
//-----------------------------------------------------------------------------
#include "SimTKsimbody.h"
using namespace SimTK;
using namespace std;
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Prototypes for local functions (functions not called by code in other files)
//-----------------------------------------------------------------------------
bool  SimulateProjectileMotion( void );
bool  WriteStringToFile(   const char outputString[], FILE *fptr )  { return fputs( outputString, fptr ) != 0; }
bool  WriteStringToScreen( const char outputString[] )              { return WriteStringToFile( outputString, stdout ); }
bool  WriteDoubleToFile( double x, int precision, FILE *fptr );
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute );
double  ConvertFromDegreesToRadians( double angleInDegrees ){return angleInDegrees/57.295779513082320876798154814105;/*1 radian is approximately 57.3 degrees*/}



//-----------------------------------------------------------------------------
// The executable program starts here
//-----------------------------------------------------------------------------
int  main( int numberOfCommandLineArguments, char *arrayOfCommandLineArguments[] )
{
   // Simulate the multibody system
   bool simulationSucceeded = SimulateProjectileMotion();

   // Keep the screen displayed until the user presses the Enter key
   WriteStringToScreen( "\n\n Press  Enter  to terminate the program: " );
   getchar();

   // The value returned by the main function is the exit status of the program.
   // A normal program exit returns 0 (other return values usually signal an error).
   return simulationSucceeded == true ? 0 : 1;
}



//-----------------------------------------------------------------------------
bool  SimulateProjectileMotion( void )
{
   // Declare a multibody system (contains one or more force and matter sub-systems)
   MultibodySystem  mbs;

   // 0. The ground's right-handed, orthogonal x,y,z unit vectors are directed with x horizontally right and y vertically upward.
   // 1. Create a gravity vector that is straight down (in the ground's frame)
   // 2. Create a uniform gravity sub-system
   // 3. Add the gravity sub-system to the multibody system
   Vec3 gravityVector( 0, -9.8, 0 );
   UniformGravitySubsystem gravity( gravityVector );
   mbs.addForceSubsystem( gravity );

   // Create a matter sub-system (the baseball)
   SimbodyMatterSubsystem  baseball;

   // Create the mass, center of mass, and inertia properties for the baseball
   const Real  massOfBaseball = 0.142;

   // The location of the baeball's center of mass is a vector from the baseball's
   // origin expressed in the "x, y, z"' unit vectors fixed in the baseball's frame.
   // Example: The vector(0,0,0) locates the baseball's center of mass at the baseball's origin.
   // Example: The vector(1,0,0) locates the baseball's center of mass 1 unit in the "x" direction from the baseball's origin.
   const Vec3  baseballCenterOfMassLocation( 0, 0, 0 );

   // Create the baseball's inertia matrix about its origin for the "x, y, z" unit vectors fixed in the baseball's frame.
   // Note: The 3x3 inertia matrix is symmetric - so only 6 elements need to be defined.
   // Ixx, Iyy, Izz  are moments  of inertia (    diagonal terms in the matrix)
   // Ixy, Ixz, Iyz  are products of inertia (off-diagonal terms in the matrix)
   const Real  Ixx = 1,  Iyy = 1,   Izz = 1;
   const Real  Ixy = 0,  Ixz = 0,   Iyz = 0;
   const Inertia  baseballInertiaMatrix( Ixx, Iyy, Izz, Ixy, Ixz, Iyz );

   // The MassProperties class holds the mass, center of mass, and inertia properties of a rigid body.
   // Although the next line creates an instance of the MassProperties class for the baeball,
   // it does not get associated with the baseball until the  addRigidBody  method.
   MassProperties  baseballMassProperties( massOfBaseball, baseballCenterOfMassLocation, baseballInertiaMatrix );

   // The motion of the baseball is related to the motion of the ground via "mobilizers"
   // The "mobilizers" specify the allowable motion of the baseball to the ground.
   // More specifically, the motion of an "outboard body" (e.g., the baseball)
   // to its inboard body (e.g., the ground) is specified by first constructing:
   // 1. An  "outboard frame"  on the ground (which hooks to the "outboard body" - the baseball)
   // 2. An  "inboard  frame"  on the baseball  (which hooks to the "inboard  body" - the ground)

   // The orientation and position of the  outboard frame  from the ground's frame is specified below.
   // The outboard frame's axes are aligned with the ground's axes and its origin is coincident with the ground's origin.
   // In other words, for this simple problem the outboard frame and the ground frame are identical.
   const Transform outboardFrameTransformFromGround;  // The default constructor is the identity transform

   // The orientation and position of the  inboard  frame  from the baseball's  frame is specified below.
   // The inboard frame's  axes are aligned with the baseball's  axes and its origin is coincident with the baseball's  origin
   // In other words, for this simple problem the inboard  frame and the baseball frame are identical.
   // Although the inboard frame can be constructed in a simple manner analogous to the outboardFrameTransformFromGround (above)
   // it is worthwhile to look at the details of the rotation matrix and position vector in the transform:
   // a. The rotation matrix relating the InboardFrame's x,y,z axes to the BaseballFrame's x,y,z axes is specified InboardFrame_BaseballFrame.
   // b. The position of the InboardFrame's origin from the BaseballFrame origin, expressed in terms of the BaseballFrame's "x, y, z" unit vectors.
   const Rotation   inboardFrameOrientationInBaseball;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       inboardFrameOriginLocationFromBaseballOrigin( 0, 0, 0 );
   const Transform  inboardFrameTransformFromBaseball( inboardFrameOrientationInBaseball, inboardFrameOriginLocationFromBaseballOrigin );

   // There are many ways that the baseball can move relative to the ground.
   // The following allows the baseball to move in "x", "y", and "z" directions.
   // Another option producing the same result is Mobilizer::Free
   Mobilizer baseballToGroundMobilizer = Mobilizer::Cartesian;
   
   // Add the properties desribed above the the basebal bodyID
   const BodyId baseballBodyId = baseball.addRigidBody( baseballMassProperties, inboardFrameTransformFromBaseball, GroundId, outboardFrameTransformFromGround, baseballToGroundMobilizer );

   // Add the matter (baseball) sub-system to the system.
   mbs.setMatterSubsystem( baseball );

   // Create a state for this system.
   // Define appropriate states for this multi-body system.
   // Set the initial time to 0.0
   State s;
   mbs.realize( s );
   s.setTime( 0.0 );

   // Set the initial values for the configuration variables (x,y,z)
   baseball.setMobilizerQ( s, baseballBodyId, 0,  0.0 );
   baseball.setMobilizerQ( s, baseballBodyId, 1,  0.0 );
   baseball.setMobilizerQ( s, baseballBodyId, 2,  0.0 );

   // Set the initial values for the motion variables
   // The baseball is hit at an angle of 35 degrees from the ground with a velocity of 44.7 m/s
   baseball.setMobilizerU( s, baseballBodyId, 0,  cos(ConvertFromDegreesToRadians(35))*44.7 );
   baseball.setMobilizerU( s, baseballBodyId, 1,  sin(ConvertFromDegreesToRadians(35))*44.7 );
   baseball.setMobilizerU( s, baseballBodyId, 2,  0.0 );

   // Create a study using the Runge Kutta Merson integrator (alternately use the CPodesIntegrator)
   RungeKuttaMerson myStudy( mbs, s );

   // Set the numerical accuracy for the integrator
   myStudy.setAccuracy( 1.0E-7 );

   // The next statement does lots of accounting
   myStudy.initialize();

   // Open a file to record the simulation results (they are also displayed on screen)
   FILE *outputFile = FileOpenWithMessageIfCannotOpen( "BaseballSimulation.txt", "w" );
   WriteStringToFile(   "time      xLocation    yLocation    xVelocity    yVelocity  mechanicalEnergy  yDifferenceFromExact\n",  outputFile );
   WriteStringToScreen( "time      xLocation    yLocation    xVelocity    yVelocity  mechanicalEnergy  yDifferenceFromExact\n" );

   // Visualize results with VTKReporter
   VTKReporter animationResults( mbs );

   // For visualization purposes only, create a red sphere with a radius of 0.5 meters (huge but visible)
   DecorativeSphere redSphere = DecorativeSphere(0.5);
   redSphere.setColor(Red);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   redSphere.setOpacity(0.0);   // 0.0 is solid and 1.0 is transparent

   // Decorate the baseball with the red sphere at the baseball's origin
   const Rotation  baseballToRedSphereOrientation; //( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3      baseballOriginToRedSphereOriginLocation( 0, 0, 0 );
   const Transform baseballToRedSphereTransform( baseballToRedSphereOrientation, baseballOriginToRedSphereOriginLocation );
   animationResults.addDecoration( baseballBodyId, baseballToRedSphereTransform, redSphere );

   // Set the numerical integration step and the time for the simulation to run
   // Run for 5.5 s, which is actually past the time the ball hits the ground, but
   // the data we are intrested in can be extracted.
   const Real dt = 0.01;
   const Real finalTime = 5.5;

   // Run the simulation and print the results
   while( 1 )
   {
      // Query for results to be printed
      Real time = s.getTime();
      Real kineticEnergy = mbs.getKineticEnergy(s);
      Real uniformGravitationalPotentialEnergy = mbs.getPotentialEnergy(s);
      Real mechanicalEnergy = kineticEnergy + uniformGravitationalPotentialEnergy;

      // Locate the baseball origin's from the ground's origin, expressed in terms of the ground's "x,y,z" unit vectors.
      // Extract the baseball's x-location and y-location from this vector.
      const Vec3 baseballLocation = baseball.calcBodyOriginLocationInBody( s, baseballBodyId, GroundId );
      Real xLocation = baseballLocation[0];
      Real yLocation = baseballLocation[1];


      // Get the baseball origin's velocity in ground, expressed in terms of the ground's "x,y,z" unit vectors.
      // Extract the baseball's y-velocity from this vector.
      const Vec3 baseballVelocity = baseball.calcBodyOriginVelocityInBody( s, baseballBodyId, GroundId );
      Real xVelocity=baseballVelocity[0];
      Real yVelocity = baseballVelocity[1];



      // Exact analytical results and difference of numerical integration results
      double yExact = 44.7*sin(ConvertFromDegreesToRadians(35))*time - 4.9*time*time;
      double yDifferenceFromExact = yExact - yLocation;

      // Print results to screen
      WriteDoubleToFile( time,                 2, stdout );
      WriteDoubleToFile( xLocation,            4, stdout );
      WriteDoubleToFile( yLocation,            4, stdout );
      WriteDoubleToFile( xVelocity,            4, stdout );
      WriteDoubleToFile( yVelocity,            4, stdout );
      WriteDoubleToFile( mechanicalEnergy,     7, stdout );
      WriteDoubleToFile( yDifferenceFromExact, 7, stdout );
      WriteStringToScreen( "\n" );
      // Print results to file with high accuracy
      WriteDoubleToFile( time,                 2, outputFile );
      WriteDoubleToFile( xLocation,            17, outputFile );
      WriteDoubleToFile( yLocation,            17, outputFile );
      WriteDoubleToFile( xVelocity,            17, outputFile );
      WriteDoubleToFile( yVelocity,            17, outputFile );
      WriteDoubleToFile( mechanicalEnergy,     7, outputFile );
      WriteDoubleToFile( yDifferenceFromExact, 17, outputFile );
      WriteStringToFile( "\n", outputFile );

      // Animate the results for this step
      animationResults.report(s);


      // Check if integration has completed
      if( time >= finalTime ) break;

      // Increment time step
      myStudy.step( time + dt);
   }
   
   fclose(outputFile);
   // Simulation completed properly
   return true;
}


//-----------------------------------------------------------------------------
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute )
{
   // Try to open the file
   FILE *Fptr1 = fopen( filename, attribute );

   // If unable to open the file, issue a message
   if( !Fptr1 )
   {
      WriteStringToScreen( "\n\n Unable to open the file: " );
      WriteStringToScreen( filename );
      WriteStringToScreen( "\n\n" );
   }

   return Fptr1;
}


//-----------------------------------------------------------------------------
bool  WriteDoubleToFile( double x, int precision, FILE *fptr )
{
   // Ensure the precision (number of digits in the mantissa after the decimal point) makes sense.
   // Next, calculate the field width so it includes one extra space to the right of the number.
   if( precision < 0 || precision > 17 ) precision = 5;
   int fieldWidth = precision + 8;

   // Create the format specifier and print the number
   char format[20];
   sprintf( format, " %%- %d.%dE", fieldWidth, precision );
   return fprintf( fptr, format, x ) >= 0;
}


